# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Command line interface for the application."""

from __future__ import annotations

import datetime
import importlib
import json
import logging
import os
import pathlib
import shlex
import shutil
import subprocess
import time
import typing as t

import click
import uvicorn

import capella_model_explorer
import capella_model_explorer.constants as c
from capella_model_explorer import app

logger = logging.getLogger(__name__)


def _get_file_timestamp(
    relpath: str | os.PathLike[str],
) -> datetime.datetime | None:
    abspath = pathlib.Path(__file__).parents[1].joinpath(relpath)
    try:
        stat = os.stat(abspath)
    except FileNotFoundError:
        return None
    return datetime.datetime.fromtimestamp(stat.st_mtime, datetime.UTC)


def _install_npm_pkgs() -> None:
    modules_ts = _get_file_timestamp("node_modules")
    if modules_ts is not None:
        maxtime = datetime.datetime.max.replace(tzinfo=datetime.UTC)
        packages_ts = min(
            _get_file_timestamp("package.json") or maxtime,
            _get_file_timestamp("pnpm-lock.yaml") or maxtime,
        )
        if packages_ts < modules_ts:
            return

    pm = _find_exe("pnpm")
    cmd = [pm, "install", "--frozen-lockfile"]
    logger.info(shlex.join(cmd))
    subprocess.check_call(cmd)


def run_container(*, log_config: dict[str, t.Any]) -> None:
    docker = _find_exe("docker")
    cmd = [
        docker,
        "run",
        "--rm",
        "-it",
        "--name=cme",
        f"-eCME_LIVE_MODE={'1' if c.LIVE_MODE else '0'}",
        f"-eCME_PORT={c.PORT}",
        f"-eCME_ROUTE_PREFIX={c.ROUTE_PREFIX}",
        f"-eCME_LOG_CONFIG={json.dumps(log_config)}",
        f"-p{c.PORT}:{c.PORT}",
    ]
    model = pathlib.Path(c.MODEL)
    if model.is_file() and model.suffix.lower() == ".aird":
        model = model.parent
    if model.is_dir():
        cmd.append("-eCME_MODEL=/model")
        cmd.append(f"-v{model.resolve()}:/model")
    else:
        cmd.append(f"-eCME_MODEL={c.MODEL}")

    templates_dir = pathlib.Path(c.TEMPLATES_DIR).resolve()
    cmd.append(f"-v{templates_dir}:/templates")

    cmd.append(c.DOCKER_IMAGE_NAME)
    logger.info(shlex.join(cmd))
    subprocess.check_call(cmd)


def run_local(*, rebuild: bool, log_config: dict[str, t.Any]) -> None:
    """Run the application locally."""
    if not pathlib.Path(c.TEMPLATES_DIR).is_dir():
        raise SystemExit(f"Templates directory not found: {c.TEMPLATES_DIR}")

    if rebuild or not pathlib.Path(c.css_bundle_path).exists():
        build_bundle(watch=False)

    logger.info("Running the application locally...")
    uvicorn.run(
        app="capella_model_explorer.app:app",
        host=c.HOST,
        port=c.PORT,
        log_config=log_config,
        reload=c.LIVE_MODE,
        reload_dirs=str(c.TEMPLATES_DIR) if c.LIVE_MODE else None,
        reload_excludes="git_askpass.py" if c.LIVE_MODE else None,
        reload_includes="*.j2" if c.LIVE_MODE else None,
    )


def run_local_dev(*, log_config: dict[str, t.Any]) -> None:
    logger.info("Running the application locally with full reload...")
    if not pathlib.Path(c.TEMPLATES_DIR).is_dir():
        raise SystemExit(f"Templates directory not found: {c.TEMPLATES_DIR}")

    bundlers = build_bundle(watch=True)
    assert bundlers is not None
    tailwind_proc, parcel_proc = bundlers
    time.sleep(1)  # avoid direct uvicorn reload when css file is written

    try:
        with tailwind_proc, parcel_proc:
            uvicorn.run(
                app="capella_model_explorer.app:app",
                host=c.HOST,
                port=c.PORT,
                log_config=log_config,
                reload=True,
                reload_dirs=".",
                reload_excludes=[
                    "git_askpass.py",
                    "input.css",
                    "compiled.css",
                ],
                reload_includes=[
                    "*.css",
                    "*.j2",
                    "*.js",
                    "*.py",
                ],
            )
    except KeyboardInterrupt:
        tailwind_proc.terminate()
        parcel_proc.terminate()


def build_bundle(
    *, watch: bool
) -> tuple[subprocess.Popen, subprocess.Popen] | None:
    """Build the script and CSS bundles."""
    logger.info("Building frontend bundles...")
    _install_npm_pkgs()

    input_css = pathlib.Path("frontend/input.css")
    if not input_css.is_file():
        raise SystemExit(f"Input CSS file not found: {input_css}")

    tailwind_cmd = [
        "pnpm",
        "exec",
        "tailwindcss",
        "--input",
        str(input_css),
        "--output",
        "frontend/compiled.css",
        *(["--watch"] if watch else []),
    ]
    parcel_cmd = [
        "pnpm",
        "exec",
        "parcel",
        "watch" if watch else "build",
        "--dist-dir=static/bundle",
        "frontend/app.js",
    ]
    if watch:
        logger.info(shlex.join(tailwind_cmd))
        logger.info(shlex.join(parcel_cmd))
        return (subprocess.Popen(tailwind_cmd), subprocess.Popen(parcel_cmd))

    logger.info(shlex.join(tailwind_cmd))
    result = subprocess.run(tailwind_cmd, check=False)
    if result.returncode != 0:
        raise SystemExit("Build failed!")

    logger.info(shlex.join(parcel_cmd))
    result = subprocess.run(parcel_cmd, check=False)
    if result.returncode != 0:
        raise SystemExit("Build failed!")

    return None


def _find_exe(name: str) -> str:
    path = shutil.which(name)
    if path is None:
        raise SystemExit(f"Cannot find {name!r}, install it and try again")
    return path


@click.group()
@click.option(
    "--log-logfmt / --log-no-logfmt",
    envvar="CME_LOGFMT",
    default=True,
    help="Use the logfmt output format for logging",
)
@click.option(
    "--log-level",
    envvar="CME_LOG_LEVEL",
    default="INFO",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        case_sensitive=False,
    ),
)
@click.option(
    "--log-file",
    envvar="CME_LOG_FILE",
    type=click.Path(dir_okay=False),
    help="Log to this file instead of stderr",
)
@click.option(
    "--log-config",
    "raw_log_config",
    envvar="CME_LOG_CONFIG",
    default=None,
    help=(
        "A JSON-encoded log config dictionary, "
        "as understood by 'logging.config.dictConfig()'. "
        "If passed, other '--log-*' arguments are ignored."
    ),
)
@click.version_option(
    version=capella_model_explorer.__version__,
    prog_name="capella-model-explorer",
    message="%(prog)s %(version)s",
)
@click.pass_context
def main(
    ctx: click.Context,
    /,
    *,
    log_logfmt: bool,
    log_level: str,
    log_file: str | None,
    raw_log_config: str | None,
) -> None:
    """Command line interface to control the application."""
    obj = ctx.ensure_object(dict)

    if raw_log_config:
        obj["log_config"] = json.loads(raw_log_config)
    else:
        if not log_file:
            handler = {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stderr",
                "formatter": "default",
                "filters": ["denoise"],
            }
        else:
            handler = {
                "class": "logging.FileHandler",
                "filename": log_file,
                "formatter": "default",
                "filters": ["denoise"],
            }

        if log_logfmt:
            formatter = {
                "()": "capella_model_explorer.core.Logfmter",
                "keys": [
                    "at",
                    "logger",
                    "msg",
                    "client_addr",
                    "method",
                    "path",
                    "http_version",
                    "status_code",
                ],
                "mapping": {
                    "at": "levelname",
                    "logger": "name",
                },
            }
        else:
            formatter = {
                "format": "%(asctime)s.%(msecs)03d <%(name)s> %(levelname)s: %(message)s",
            }

        obj["log_config"] = {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {"default": formatter},
            "filters": {
                "denoise": {
                    "()": "capella_model_explorer.core.SuppressWebsocketNoise",
                }
            },
            "handlers": {"default": handler},
            "loggers": {
                "": {"level": log_level, "handlers": ["default"]},
                "uvicorn": {"level": "NOTSET", "propagate": True},
                "uvicorn.error": {"level": "NOTSET", "propagate": True},
                "uvicorn.access": {"level": "NOTSET", "propagate": True},
            },
        }


@main.command()
@click.option(
    "--container",
    is_flag=True,
    default=False,
    show_default=True,
    help="Launch as a Docker container.",
)
@click.option(
    "--dev",
    is_flag=True,
    show_default=True,
    help="Launch in development mode with full auto-reload."
    " This option cannot be used together with --container.",
)
@click.option(
    "-h",
    "--host",
    envvar="CME_HOST",
    default=c.Defaults.host,
    show_default=True,
    help="The hostname or IP address to bind to."
    " Ignored when running with '--container'.",
)
@click.option(
    "-p",
    "--port",
    envvar="CME_PORT",
    default=c.Defaults.port,
    show_default=True,
    help="The port to listen on.",
)
@click.option(
    "-m",
    "--model",
    envvar="CME_MODEL",
    default=c.Defaults.model,
    show_default=True,
    help="The Capella model to load (file, URL or JSON string).",
)
@click.option(
    "-t",
    "--templates-dir",
    envvar="CME_TEMPLATES_DIR",
    default=str(c.Defaults.templates_dir.resolve()),
    show_default=True,
    help="The directory containing the templates.",
)
@click.option(
    "--live-mode/--no-live-mode",
    envvar="CME_LIVE_MODE",
    default=c.Defaults.live_mode,
    show_default=True,
    help=(
        "Control automatic reloading of templates on changes. "
        "Ignored in '--dev' mode, where it is always enabled."
    ),
)
@click.option(
    "--route-prefix",
    envvar="CME_ROUTE_PREFIX",
    default="",
    show_default=True,
    help="Add a prefix to all web routes."
    f" (Note: this prefix does not apply to '{app.metrics.to()}').",
)
@click.option(
    "--image",
    envvar="CME_DOCKER_IMAGE_NAME",
    default=c.Defaults.docker_image_name,
    show_default=True,
    help="The Docker image to use with '--container'.",
)
@click.option(
    "--skip-rebuild/--no-skip-rebuild",
    default=False,
    help="Don't rebuild already existing assets.",
)
@click.option(
    "--debug-spinner",
    envvar="CME_DEBUG_SPINNER",
    is_flag=True,
    default=False,
    help="Keep the template loading spinner spinning until clicked",
)
@click.pass_context
def run(
    ctx: click.Context,
    /,
    *,
    container: bool,
    dev: bool,
    host: str,
    port: int,
    model: str,
    templates_dir: str,
    live_mode: bool,
    route_prefix: str,
    image: str,
    skip_rebuild: bool,
    debug_spinner: bool,
) -> None:
    """Run the application."""
    os.environ["CME_HOST"] = host
    os.environ["CME_PORT"] = str(port)
    os.environ["CME_MODEL"] = model
    os.environ["CME_TEMPLATES_DIR"] = templates_dir
    os.environ["CME_LIVE_MODE"] = "1" if live_mode else "0"
    os.environ["CME_ROUTE_PREFIX"] = route_prefix
    os.environ["CME_DOCKER_IMAGE_NAME"] = image
    os.environ["CME_DEBUG_SPINNER"] = "01"[debug_spinner]
    importlib.reload(c)

    if container and dev:
        raise click.UsageError(
            "Options --container and --dev are mutually exclusive."
        )

    if container:
        run_container(log_config=ctx.obj["log_config"])
    elif dev:
        run_local_dev(log_config=ctx.obj["log_config"])
    else:
        run_local(rebuild=not skip_rebuild, log_config=ctx.obj["log_config"])


@main.command()
@click.option(
    "--watch",
    is_flag=True,
    default=False,
    show_default=True,
    help="Watch for changes and rebuild automatically.",
)
def build(*, watch: bool) -> None:
    """Build the frontend script and style bundles."""
    procs = build_bundle(watch=watch)
    if procs is not None:
        tailwind, parcel = procs
        with tailwind, parcel:
            pass


@main.command("pre-commit-setup")
def pre_commit_setup_cmd() -> None:
    """Install tools needed for pre-commit hooks."""
    pre_commit = _find_exe("pre-commit")
    logger.info("Installing tools needed for pre-commit hooks...")
    _install_npm_pkgs()
    logger.info("Installing pre-commit hooks...")
    subprocess.check_call([pre_commit, "install-hooks"])
    subprocess.check_call([pre_commit, "install"])


if __name__ == "__main__":
    main()



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import base64
import importlib.resources as imr
import json
import typing as t

from fasthtml import ft, svg

import capella_model_explorer
from capella_model_explorer import app, icons, reports, state
from capella_model_explorer import constants as c

GITHUB_URL = "https://github.com/DSD-DBS/capella-model-explorer"


def application_shell(
    *content: t.Any,
    template: reports.Template | None,
    element: str | None,
) -> tuple[ft.Title, ft.Main]:
    return (
        ft.Title(f"{state.model.name} - Model Explorer"),
        ft.Body(
            navbar(template, element),
            ft.Main(
                *content,
                id="root",
                cls=(
                    "grow",
                    "w-full",
                    "overflow-y-auto",
                ),
            ),
            # placeholder for script injection per outerHTML swap
            ft.Script(id="script"),
            cls=(
                "bg-neutral-100",
                "dark:bg-neutral-900",
                "flex",
                "flex-col",
                "h-screen",
                "min-h-screen",
                "print:h-auto",
                "print:min-h-auto",
            ),
            hx_ext="morph",
        ),
    )


def breadcrumb(*, label: str, url: str) -> ft.Li:
    return (
        ft.Li(
            ft.Div(
                svg.Svg(
                    svg.Path(d="M.293 0l22 22-22 22h1.414l22-22-22-22H.293z"),
                    viewbox="0 0 24 44",
                    preserveaspectratio="none",
                    aria_hidden="true",
                    cls="h-full w-6 stroke-1 stroke-primary-400 dark:stroke-neutral-700",
                ),
                ft.A(
                    label,
                    href=url,
                    cls=(
                        "dark:hover:text-neutral-100",
                        "dark:text-neutral-400",
                        "font-medium",
                        "ml-4",
                        "hover:text-neutral-50",
                        "text-neutral-300",
                        "text-sm",
                    ),
                ),
                cls="flex items-center",
            ),
            cls="flex",
        ),
    )


def breadcrumbs(
    template: reports.Template | None = None,
    element_id: str | None = None,
    /,
    *,
    oob: bool = False,
) -> ft.Nav:
    components = []

    if template is not None:
        components.append(
            ft.Li(
                ft.Div(
                    ft.A(
                        icons.home(),
                        ft.Span("Home", cls="sr-only"),
                        href=app.app.url_path_for("main_home"),
                        hx_get=app.app.url_path_for("main_home"),
                        hx_target="#root",
                        hx_push_url="true",
                    ),
                    cls="flex items-center",
                ),
                cls="flex",
            )
        )

        url = app.app.url_path_for("template_page", template_id=template.id)
        components.append(breadcrumb(label=template.name, url=url))

    if element_id is not None:
        assert template is not None, "Model element passed without template"
        element = state.model.by_uuid(element_id)
        url = app.app.url_path_for(
            "template_page",
            template_id=template.id,
            model_element_uuid=element_id,
        )
        components.append(breadcrumb(label=element.name, url=url))

    return ft.Nav(
        ft.Ol(
            *components,
            role="list",
            cls="flex space-x-4 rounded-md px-6",
        ),
        id="breadcrumbs",
        aria_label="Breadcrumb",
        cls="flex grow",
        **({"hx_swap_oob": "true"} if oob else {}),
    )


def model_information() -> ft.Div:
    """Render the model information including the badge."""
    badge = "data:image/svg+xml;base64," + base64.standard_b64encode(
        state.model.description_badge.encode("utf-8")
    ).decode("ascii")
    return ft.Div(
        ft.H1(state.model.name, cls="text-xl"),
        ft.P(
            f"Capella version: {state.model.info.capella_version}",
        ),
        ft.Img(
            src=badge,
            alt="Model description badge",
            cls="object-scale-down",
        ),
        cls=(
            "bg-white",
            "dark:bg-neutral-800/50",
            "dark:shadow-none",
            "dark:text-neutral-400",
            "flex",
            "flex-col",
            "max-w-full",
            "m-2",
            "mt-4",
            "p-4",
            "place-items-center",
            "rounded-lg",
            "shadow-primary-500",
            "shadow-xs",
            "svg-display",
            "text-center",
            "text-neutral-700",
        ),
    )


def model_object_button(
    *, template: reports.Template, model_element: dict, selected: bool = False
) -> ft.A:
    return ft.A(
        ft.Div(model_element["name"]),
        ft.Div(model_element["uuid"], cls="text-xs text-sky-700")
        if state.show_uuids
        else None,
        id=f"model-element-{model_element['uuid']}",
        aria_selected="true" if selected else "false",
        cls=(
            "aria-selected:bg-primary-500",
            "aria-selected:dark:hover:bg-primary-500",
            "aria-selected:dark:text-white",
            "aria-selected:text-white",
            "bg-neutral-200",
            "break-words",
            "cursor-pointer",
            "dark:bg-neutral-800",
            "dark:hover:bg-neutral-700",
            "dark:text-neutral-400",
            "duration-300",
            "flex",
            "flex-col",
            "hover:bg-primary-50",
            "hover:scale-[1.02]",
            "model_object_btn",
            "p-4",
            "place-items-left",
            "rounded-md",
            "shadow-sm",
            "text-left",
            "transition",
            "w-full",
        ),
        hx_trigger="click",
        hx_get=app.app.url_path_for(
            "template_page",
            template_id=template.id,
            model_element_uuid=model_element["uuid"],
        ),
        hx_push_url="true",
        hx_include='[name="search"]',
        hx_target="#template_container",
    )


def model_elements_list(
    *,
    template: reports.Template,
    selected_id: str | None,
    search: str = "",
) -> ft.Div:
    search_words = search.lower().split()
    model_elements = [
        obj
        for obj in sorted(template.instances, key=lambda x: x["name"])
        if (n := obj["name"].lower()) and all(w in n for w in search_words)
    ]
    return ft.Div(
        ft.Div(
            *(
                model_object_button(
                    template=template,
                    model_element=model_element,
                    selected=model_element["uuid"] == selected_id,
                )
                for model_element in model_elements
            ),
            cls="flex flex-col space-y-4 pl-2 pr-4 my-2",
        ),
        id="model_object_list",
        cls="overflow-auto grow",
    )


def navbar(template: reports.Template | None, element: str | None) -> ft.Nav:
    return ft.Nav(
        breadcrumbs(template, element),
        ft.Button(
            icons.printer(),
            onclick="window.print();",
            id="print-button",
            title="Print report",
            cls="hidden cursor-pointer",
        ),
        ft.Button(
            icons.toc_icon(),
            id="toc-toggle-button",
            title="Toggle table of contents",
            onclick="toggleToc()",
            cls="hidden xl:hidden cursor-pointer",
        ),
        ft.Button(
            icons.theme_system(id="dark-mode-icon-system"),
            icons.theme_dark(id="dark-mode-icon-dark", cls=("hidden",)),
            icons.theme_light(id="dark-mode-icon-light", cls=("hidden",)),
            id="dark-mode-button",
            title="Toggle dark mode",
            cls="cursor-pointer",
        ),
        id="page-header",
        cls=(
            "bg-primary-500",
            "dark:bg-neutral-800",
            "dark:border-b",
            "dark:border-neutral-700",
            "flex",
            "gap-2",
            "h-12",
            "min-h-12",
            "max-h-12",
            "px-4",
            "py-0",
            "print:hidden",
            "sticky",
            "top-0",
            "w-full",
            "z-50",
        ),
    )


def report_placeholder(
    template: reports.Template | None,
    model_element_uuid: str | None,
) -> t.Any:
    if template is None:
        ph_content = ft.Div(
            ft.Span(
                "Select a model element.",
                cls="text-slate-600 dark:text-slate-500 p-4 italic m-auto",
            ),
            cls="flex justify-center place-items-center h-full w-full",
        )

    else:
        render_environment = reports.compute_cache_key(template)
        headers = json.dumps({"Render-Environment": render_environment})

        ph_content = ft.Div(
            icons.spinner(),
            hx_trigger="click" if c.DEBUG_SPINNER else "load",
            hx_get=app.rendered_report.to(
                template_id=template.id,
                model_element_uuid=model_element_uuid,
            ),
            hx_headers=headers,
            hx_target="#template_container",
            cls="flex justify-center place-items-center h-full w-full",
        )

    return ph_content


def reports_page() -> ft.Div:
    return ft.Div(
        *(template_category(t) for t in state.template_categories),
        id="reports-index",
        cls=(
            "bg-white",
            "dark:bg-neutral-800/50",
            "dark:shadow-none",
            "flex",
            "flex-col",
            "m-2",
            "p-4",
            "place-items-center",
            "rounded-lg",
            "shadow-primary-500",
            "shadow-xs",
            "space-y-16",
        ),
    )


def search_field(template: reports.Template, search: str) -> ft.Div:
    search_field_threshold = 3
    return ft.Div(
        ft.Input(
            icons.magnifying_glass(),
            type="search",
            id="instance-search",
            name="search",
            placeholder="Search",
            value=search,
            cls=(
                "-outline-offset-1",
                "bg-white",
                "block",
                "col-start-1",
                "dark:bg-neutral-800",
                "dark:focus:outline-white",
                "dark:text-neutral-400",
                "focus:-outline-offset-2",
                "focus:outline-2",
                "focus:outline-primary-500",
                "grow",
                "outline",
                "outline-1",
                "outline-neutral-300",
                "pl-8",
                "placeholder:text-neutral-400",
                "pr-3",
                "py-1.5",
                "rounded-md",
                "row-start-1",
                "text-neutral-900",
            ),
            hx_trigger="input changed delay:20ms, search",
            hx_get=app.model_object_list.to(
                template_id=template.id,
                selected_model_element_uuid="",
            ),
            hx_swap="outerHTML",
            hx_target="#model_object_list",
            hx_preserve="true",
            autofocus="true",
        ),
        cls=(
            "hidden"
            if template.instance_count <= search_field_threshold
            else "",
            "grid",
            "grid-cols-1",
            "pl-2",
            "pr-4",
        ),
    )


def template_card(template: reports.Template) -> ft.A:
    url = app.app.url_path_for("template_page", template_id=template.id)

    chips = []
    if template.isExperimental:
        c = ft.Div(
            icons.badge_experimental(),
            ft.P("Experimental"),
            cls=(
                "bg-yellow-100",
                "border",
                "border-yellow-800",
                "dark:border-yellow-600",
                "dark:bg-yellow-900",
                "dark:text-yellow-300",
                "flex",
                "flex-row",
                "font-medium",
                "max-h-7",
                "me-2",
                "px-2.5",
                "py-1",
                "rounded-full",
                "space-x-2",
                "text-xs",
                "text-yellow-800",
            ),
        )
        chips.append(c)

    if template.isStable:
        c = ft.Div(
            icons.badge_stable(),
            ft.P("Stable"),
            cls=(
                "bg-green-100",
                "border",
                "border-green-800",
                "dark:bg-green-700",
                "dark:border-green-500",
                "dark:text-green-200",
                "flex",
                "flex-row",
                "font-medium",
                "max-h-7",
                "me-2",
                "px-2.5",
                "py-1",
                "rounded-full",
                "space-x-2",
                "text-green-800",
                "text-xs",
            ),
        )
        chips.append(c)

    if template.isDocument:
        c = ft.Div(
            icons.badge_document(),
            ft.P("Document"),
            cls=(
                "bg-blue-200",
                "border",
                "border-blue-800",
                "dark:bg-blue-700",
                "dark:border-blue-400",
                "dark:text-blue-200",
                "flex",
                "flex-row",
                "font-medium",
                "max-h-7",
                "me-2",
                "px-2.5",
                "py-1",
                "rounded-full",
                "space-x-2",
                "text-blue-800",
                "text-xs",
            ),
        )
        chips.append(c)

    if chips:
        chip_container = ft.Div(
            *chips, cls="flex flex-row grow place-items-end p-4"
        )
    else:
        chip_container = None

    return ft.A(
        ft.Div(
            ft.Div(
                template.name,
                cls="text-neutral-100 dark:text-neutral-400 text-2xl",
            ),
            ft.Div(
                icons.file_stack()
                if template.instance_count > 1
                else icons.report(),
                ft.Span(
                    template.instance_count,
                    cls="hidden" if template.instance_count <= 1 else "block",
                ),
                cls=(
                    "dark:text-neutral-400",
                    "flex",
                    "flex-row",
                    "items-top",
                    "pt-2",
                    "space-x-2",
                    "text-primary-50",
                ),
            ),
            cls=(
                "_template-card-header",
                "bg-primary-500",
                "dark:bg-neutral-900",
                "flex",
                "flex-row",
                "justify-between",
                "p-4",
                "rounded-t-lg",
            ),
        ),
        ft.P(
            template.description,
            cls=(
                "_template-card-description",
                "dark:text-neutral-300",
                "p-4",
                "text-left",
                "text-neutral-800",
            ),
        ),
        chip_container,
        cls=(
            "_template-card",
            "active:border-blue-600",
            "bg-white",
            "dark:bg-neutral-800",
            "dark:hover:bg-neutral-700",
            "dark:shadow-neutral-900",
            "dark:shadow-xs",
            "duration-300",
            "flex",
            "flex-col",
            "hover:bg-primary-50",
            "hover:cursor-pointer",
            "hover:scale-105",
            "m-4",
            "rounded-lg",
            "shadow-lg",
            "transition",
            "w-80",
        ),
        href=url,
        hx_get=url,
        hx_target="#root",
        hx_push_url="true",
    )


def template_category(
    template_category: reports.TemplateCategory,
) -> ft.Div:
    return ft.Div(
        ft.P(
            f"{template_category.idx} Reports",
            cls=(
                "border-b",
                "border-neutral-900",
                "dark:border-neutral-700",
                "dark:text-neutral-400",
                "font-semibold",
                "pb-2",
                "text-2xl",
                "text-center",
                "text-neutral-800",
            ),
        ),
        ft.Div(
            *[
                template_card(template)
                for template in template_category.templates
            ],
            cls=(
                "2xl:grid-cols-4",
                "auto-rows-auto",
                "content-around",
                "grid",
                "grid-cols-1 lg:grid-cols-2 xl:grid-cols-3",
                "grid-flow-row",
                "pt-4",
            ),
        ),
        id="template-category",
        cls="rounded-lg w-full my-2 mx-auto justify-center",
    )


def template_container(content: t.Any) -> ft.Div:
    return ft.Div(
        content,
        id="template_container",
        cls=(
            "bg-white",
            "dark:bg-neutral-800",
            "dark:border-b",
            "dark:lg:border-l",
            "dark:border-neutral-700",
            "dark:shadow-neutral-700",
            "min-h-full",
            "html-content",
            "flex",
            "items-start",
            "justify-center",
            "p-4",
            "print:bg-white",
            "print:m-0",
            "print:ml-6",
            "print:p-0",
            "svg-display",
            "template-container",
            "w-full",
        ),
    )


def template_sidebar(
    *,
    template: reports.Template,
    selected_model_element_uuid: str | None,
    search: str = "",
    oob: bool = False,
) -> ft.Div:
    sidebar_caption = (
        ft.Div(
            ft.H1(
                template.name,
                cls="text-xl text-neutral-700 dark:text-neutral-400",
            ),
            ft.H2(
                template.description,
                cls="text-sm text-neutral-900 dark:text-neutral-400",
            ),
            cls="mx-2",
        ),
    )
    return ft.Div(
        sidebar_caption,
        search_field(template, search=search),
        model_elements_list(
            template=template,
            selected_id=selected_model_element_uuid,
            search=search,
        ),
        id="template-sidebar",
        cls=(
            "dark:bg-neutral-900",
            "flex",
            "flex-col",
            "h-full",
            "lg:max-h-[calc(100vh-12*var(--spacing))]",
            "lg:w-96",
            "max-h-[calc(0.85*(100vh-12*var(--spacing)))]",
            "pl-4",
            "print:hidden",
            "py-4",
            "rounded-lg",
            "space-y-4",
            "sticky",
            "top-0",
        ),
        hx_swap_oob=oob and "morph",
    )


def table_of_contents(toc_items: list[dict]) -> ft.Div:
    if not toc_items:
        return ft.Div()

    return ft.Div(
        ft.Div(
            ft.H3(
                "Table of Contents",
                cls=(
                    "text-lg",
                    "font-semibold",
                    "text-neutral-800",
                    "dark:text-neutral-300",
                    "mb-3",
                    "px-3",
                    "pt-4",
                    "xl:pt-0",
                ),
            ),
            ft.Nav(
                ft.Ul(
                    *(toc_item(item) for item in toc_items),
                    cls="space-y-1",
                ),
                cls="overflow-y-auto xl:max-h-[calc(100vh-16rem)]",
            ),
            cls="xl:sticky xl:top-4 h-full xl:h-auto",
        ),
        ft.Script(
            imr.read_text(__name__.rsplit(".", 1)[0], "table-of-contents.js")
        ),
        id="table-of-contents",
        cls=(
            "fixed",
            "top-12",
            "right-0",
            "bottom-0",
            "w-80",
            "bg-white",
            "dark:bg-neutral-900",
            "border-l",
            "border-neutral-300",
            "dark:border-neutral-700",
            "shadow-lg",
            "transform",
            "translate-x-full",
            "xl:translate-x-0",
            "transition-transform",
            "duration-300",
            "z-40",
            "overflow-y-auto",
            "pl-4",
            "xl:relative",
            "xl:transform-none",
            "xl:shadow-none",
            "xl:border-none",
            "xl:bg-transparent",
            "xl:dark:bg-transparent",
            "xl:w-80",
            "xl:top-0",
            "xl:overflow-visible",
            "xl:pl-6",
            "shrink-0",
            "print:hidden",
        ),
    )


def toc_item(item: dict) -> ft.Li:
    indent_map = {
        1: "ml-0",
        2: "ml-3",
        3: "ml-6",
        4: "ml-9",
        5: "ml-12",
        6: "ml-15",
    }
    return ft.Li(
        ft.A(
            item["text"],
            href=f"#{item['id']}",
            cls=(
                "block",
                "text-sm",
                "text-neutral-700",
                "dark:text-neutral-400",
                "hover:text-primary-600",
                "dark:hover:text-primary-400",
                "py-1.5",
                "rounded",
                "transition-colors",
                "duration-300",
                "toc-link",
                indent_map[item["level"]],
            ),
            data_target=item["id"],
        ),
    )


def bottom_bar() -> ft.Div:
    """Return container for bottom bar."""
    version = capella_model_explorer.__version__
    if "dev" in version:
        version_element = ft.Span(
            f"Capella-Model-Explorer: v{version}", cls="dark:text-gray-300"
        )
    else:
        version_element = ft.A(
            ft.Span(f"Capella-Model-Explorer: v{version}"),
            href=f"{GITHUB_URL}/releases/v{version}",
            target="_blank",
            cls="hover:underline dark:text-gray-300",
        )

    return ft.Div(
        ft.Div(version_element),
        ft.Div(
            ft.A(
                ft.Span("Contribute on GitHub"),
                icons.github_logo(),
                href=GITHUB_URL,
                target="_blank",
                cls="flex items-center gap-2 hover:underline dark:text-gray-300",
            ),
            cls="md:w-fit md:mx-auto",
        ),
        cls="md:grid md:grid-cols-3 w-full px-2 pb-2",
    )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import dataclasses
import importlib.resources as imr
import pathlib
import sys
import typing as t
import uuid

import fasthtml.starlette
from fasthtml import common as fh
from fasthtml import ft

from capella_model_explorer import core

LAUNCH_ID = str(uuid.uuid4())
"""A unique ID that identifies this launch of the model explorer.

This ID is used to invalidate client-side caches when re-launching the
server, in cases where we can't reliably ensure that the model hasn't
changed between launches. Examples include models loaded from local
files (not backed by a Git repository) or remote plain HTTP servers.
"""

CACHE_MAX_AGE = 365 * 24 * 60 * 60  # 1 year

CONFIG = fasthtml.starlette.Config(env_prefix="CME_")


@dataclasses.dataclass
class Defaults:
    docker_image_name: t.Final[str] = "capella-model-explorer:latest"
    host: t.Final[str] = "0.0.0.0"
    live_mode: bool = True
    model: t.Final[str] = (
        "git+https://github.com/DSD-DBS/Capella-IFE-sample.git"
    )
    port: t.Final[int] = 8000
    primary_color_hue: t.Final[int] = 231
    route_prefix: t.Final[str] = ""
    templates_dir: t.Final[pathlib.Path] = pathlib.Path("templates")


DEBUG_SPINNER = CONFIG("DEBUG_SPINNER", cast=bool, default=False)

DOCKER_IMAGE_NAME: str = CONFIG(
    "DOCKER_IMAGE_NAME", default=Defaults.docker_image_name
)
HOST: str = CONFIG("HOST", default=Defaults.host)
MODEL: str = CONFIG("MODEL", default=Defaults.model)
PORT: int = CONFIG("PORT", cast=int, default=Defaults.port)
PRIMARY_COLOR_HUE: int = CONFIG(
    "PRIMARY_COLOR_HUE", cast=int, default=Defaults.primary_color_hue
)
LIVE_MODE: t.Final[bool] = CONFIG(
    "LIVE_MODE", cast=bool, default=Defaults.live_mode
)

ROUTE_PREFIX: t.Final[str] = CONFIG(
    "ROUTE_PREFIX", default=Defaults.route_prefix
).rstrip("/")
TEMPLATES_DIR: t.Final[pathlib.Path] = CONFIG(
    "TEMPLATES_DIR",
    cast=pathlib.Path,
    default=Defaults.templates_dir.resolve(),
).resolve()

css_bundle_path = "static/bundle/app.css"
css_bundle_hash = core.compute_file_hash(css_bundle_path)

favicon_path = "static/favicon.svg"
favicon_hash = core.compute_file_hash(favicon_path)

js_bundle_path = "static/bundle/app.js"
js_bundle_hash = core.compute_file_hash(js_bundle_path)

HEADERS: t.Final[list[fh.Link | ft.Script]] = [
    fh.Link(
        rel="stylesheet",
        href=f"{ROUTE_PREFIX}/{css_bundle_path}?v={css_bundle_hash}",
        type="text/css",
    ),
    fh.Link(
        rel="icon",
        href=f"{ROUTE_PREFIX}/{favicon_path}?v={favicon_hash}",
        type="image/x-icon",
    ),
    ft.Script(
        charset="utf-8",
        src=f"{ROUTE_PREFIX}/{js_bundle_path}?v={js_bundle_hash}",
    ),
    fh.HighlightJS(langs=["python"]),
    ft.Script(
        imr.read_text(__name__.rsplit(".", 1)[0], "inline-dark-mode.js")
    ),
    ft.Style(
        f":root {{ --primary-color-hue: {PRIMARY_COLOR_HUE}; }}\n"
        ":not(:defined) { visibility: hidden; }"
    ),
]
if sys.stderr.isatty():
    LOGGING_COLORS = {
        "CRITICAL": "\x1b[1;31m",  # Bold and red
        "ERROR": "\x1b[31m",  # red
        "WARNING": "\x1b[33m",  # yellow
        "INFO": "\x1b[32m",  # green
        "DEBUG": "\x1b[3;90m",  # Italic and dark gray
        "RESET": "\x1b[m",
    }
else:
    LOGGING_COLORS = {
        "CRITICAL": "",
        "ERROR": "",
        "WARNING": "",
        "INFO": "",
        "DEBUG": "",
        "RESET": "",
    }



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import base64
import json
import logging
import operator
import pathlib
import re
import traceback
import typing as t

import capellambse
import capellambse.model as m
import capellambse_context_diagrams
import jinja2
import markupsafe
import pydantic as p
import yaml
from lxml import etree
from lxml import html as lxml_html

import capella_model_explorer
import capella_model_explorer.constants as c
from capella_model_explorer import app, core, state

SVG_PLACEHOLDER_MARKUP = markupsafe.Markup(
    '<div class="svg-container relative inline-block cursor-wait px-6 py-4 animate-pulse'
    ' rounded-full bg-primary-500 text-neutral-300 dark:bg-primary-700 dark:text-neutral-400"'
    ' hx-trigger="load" hx-get="{url}" hx-swap="outerHTML" hx-headers="{headers}">'
    "Loading diagram..."
    "</div>"
)

SVG_ERROR_MARKUP = markupsafe.Markup(
    '<div class="svg-container relative inline-block px-6 py-4 w-full rounded-3xl bg-red-500'
    ' text-neutral-300 dark:bg-red-700 dark:text-neutral-400">'
    '<span class="cursor-pointer w-full"'
    " hx-on:click=\"this.nextElementSibling.classList.toggle('hidden')\">"
    "Error rendering diagram (click for details)"
    "</span>"
    '<div class="overflow-auto w-full hidden">'
    '<pre class="mt-2 p-2 text-xs bg-red-800 text-white rounded-md overflow-auto w-full">{traceback}</pre>'
    "</div></div>"
)

SVG_WRAP_MARKUP = markupsafe.Markup(
    '<div class="svg-container relative inline-block cursor-pointer"'
    ' onclick="openDiagramViewer(this)" data-img="{svg_data}">'
    '<img src="{svg_data}" alt="{title}"'
    ' class="transition hover:scale-105 hover:shadow-lg hover:shadow-normal-500/50">'
    "</div>"
)

logger = logging.getLogger(__name__)


class Template(p.BaseModel):
    id: str = p.Field(title="Template identifier")
    name: str = p.Field(title="Template name")
    category: str = p.Field(title="Template category")
    description: str = p.Field(title="Template description")

    scope: TemplateScope | None = p.Field(default=None, title="Template scope")
    single: bool | None = p.Field(None, title="Single instance flag")
    isStable: bool | None = p.Field(None, title="Stable template flag")
    isDocument: bool | None = p.Field(None, title="Document template flag")
    isExperimental: bool | None = p.Field(
        None, title="Experimental template flag"
    )
    path: pathlib.Path = p.Field(title="Absolute file path to template")
    error: str | None = p.Field(None, title="Broken template flag")
    traceback: str | None = p.Field(None, title="Template error traceback")
    instance_count: int = p.Field(0, title="Number of instances")
    instances: list[dict] = p.Field([], title="List of instances")

    def model_post_init(self, _):
        self._compute_instances()

    def _compute_instances(self) -> None:
        if self.single:
            self.instance_count = 1
            return
        try:
            if self.scope is not None:
                self.instances = [
                    self._simple_object(obj)
                    for obj in self._find_objects(
                        obj_type=self.scope.type,
                        below=self.scope.below,
                        attr=None,
                        filters=self.scope.filters,
                    )
                ]
            self.instance_count = len(self.instances)
        except Exception as e:
            self.error = f"Template scope error: {e}"
            self.traceback = traceback.format_exc()

    def _find_objects(
        self, obj_type=None, below=None, attr=None, filters=None
    ):
        if attr:
            getter = operator.attrgetter(attr)
            objects = getter(state.model)
            if hasattr(objects, "_element"):
                objects = [objects]
            elif not isinstance(objects, m.ElementList):
                raise ValueError(
                    f"Expected a list of model objects or a single model object "
                    f"for {attr!r} of the model, got {objects!r}"
                )
        elif below and obj_type:
            getter = operator.attrgetter(below)
            objects = state.model.search(obj_type, below=getter(state.model))
        elif obj_type:
            objects = state.model.search(obj_type)
        else:
            raise ValueError("No search criteria provided")

        if filters:
            filtered = []
            for object in objects:
                for attr_key, filter in filters.items():
                    attr = getattr(object, attr_key)
                    if filter == "not_empty":
                        if attr:
                            filtered.append(object)
                    elif attr == filter:
                        filtered.append(object)
            objects = filtered
        return objects

    def _simple_object(self, obj) -> dict[str, t.Any]:
        if not obj:
            return {}
        if obj.name:
            name = obj.name
        else:
            name = obj.long_name if hasattr(obj, "long_name") else "undefined"

        return {"uuid": obj.uuid, "name": str(name)}


class TemplateScope(p.BaseModel):
    type: str | None = p.Field(None, title="Model Element Type")
    below: t.Literal["oa", "sa", "la", "pa"] | None = p.Field(
        None, title="Model element to search below, scope limiter"
    )
    filters: dict[str, t.Any] | None = p.Field(
        {}, title="Filters to apply to the search"
    )

    def applies_to(self, obj: m.ModelElement | m.AbstractDiagram) -> bool:
        if type(obj).__name__ != self.type:
            return False

        if (
            (self.below == "oa" and obj.layer != obj._model.oa)
            or (self.below == "sa" and obj.layer != obj._model.sa)
            or (self.below == "la" and obj.layer != obj._model.la)
            or (self.below == "pa" and obj.layer != obj._model.pa)
        ):
            return False

        for key, value in (self.filters or {}).items():
            try:
                actual = getattr(obj, key)
            except AttributeError:
                return False
            if actual != value:
                return False

        return True


class TemplateCategory(p.BaseModel):
    idx: str = p.Field(title="Category Identifier")
    templates: list[Template] = p.Field(
        default_factory=list, title="Templates in this category"
    )


def template_by_id(id_: str) -> Template | None:
    for template in state.templates:
        if template.id == id_:
            return template
    return None


def _register_template_category(category: str) -> None:
    state.template_categories.append(
        TemplateCategory(
            idx=category,
            templates=[t for t in state.templates if t.category == category],
        )
    )


def _make_href(
    obj: m.ModelElement | m.AbstractDiagram,
) -> str | None:
    generic_template: Template | None = None
    for template in state.templates:
        if template.path.name.startswith("__generic__"):
            generic_template = template
            continue

        if template.single:
            continue

        if template.scope and not template.scope.applies_to(obj):
            continue

        return app.app.url_path_for(
            "template_page",
            template_id=template.id,
            model_element_uuid=obj.uuid,
        )

    if generic_template is None:
        return None
    return app.app.url_path_for(
        "template_page",
        template_id=generic_template.id,
        model_element_uuid=obj.uuid,
    )


def finalize(markup: t.Any) -> object:
    if markup is None:
        return ""

    if isinstance(markup, m.AbstractDiagram):
        logger.warning(
            (
                "Synchronously rendering diagram inline, which is slow - "
                "please update the template to use 'render_diagram()' instead: "
                "%r (%s)"
            ),
            markup.name,
            markup.uuid,
        )
        return render_diagram(markup.render("svg"), markup.name)

    if isinstance(markup, capellambse.diagram.Diagram):
        logger.warning(
            (
                "Placing synchronously rendered diagram inline, which is slow - "
                "please update the template to use 'render_diagram()' instead: "
                "%r (%s)"
            ),
            markup.name,
            markup.uuid,
        )
        svg = m.diagram.convert_format(None, "svg", markup)
        return render_diagram(svg, markup.name)

    markup = markupsafe.escape(markup)
    return capellambse.helpers.replace_hlinks(markup, state.model, _make_href)


def make_href_filter(obj: object) -> str | None:
    if jinja2.is_undefined(obj) or obj is None:
        return "#"

    if isinstance(obj, m.ElementList):
        raise TypeError("Cannot make an href to a list of elements")
    if not isinstance(obj, m.ModelElement | m.AbstractDiagram):
        raise TypeError(f"Expected a model object, got {obj!r}")
    try:
        state.model.by_uuid(obj.uuid)
    except KeyError:
        return "#"
    return _make_href(obj)


def diagram_placeholder(
    parent: capellambse.model.ModelElement,
    attr: str,
    /,
    **kw: t.Any,
) -> t.Any:
    logger.info(
        "Generating diagram placeholder for %r on %r", attr, parent.uuid
    )
    if not isinstance(parent, capellambse.model.ModelElement):
        raise TypeError("Parent object must be a model element")
    if not isinstance(attr, str) or not attr.isidentifier():
        raise TypeError("Attribute must be a Python identifier string")
    try:
        can_find = state.model.by_uuid(parent.uuid) == parent
    except KeyError:
        can_find = False
    if not can_find:
        raise TypeError(f"Cannot find object by its UUID: {parent!r}")

    try:
        diag = getattr(parent, attr)
    except AttributeError:
        raise TypeError(
            f"Parent object ({type(parent).__name__}) "
            f"does not have an attribute {attr!r}"
        ) from None
    if not isinstance(diag, capellambse.model.AbstractDiagram):
        raise TypeError(
            f"Expected a diagram at {type(parent).__name__}.{attr}, "
            f"got {type(diag).__name__}"
        )

    url = app.app.url_path_for("render_diagram", parent=parent.uuid, attr=attr)
    if kw:
        try:
            params = json.dumps(kw)
        except Exception:
            raise TypeError(
                f"Render parameters must be JSON serializable: {kw!r}"
            ) from None
        url = f"{url}?params={params}"

    render_environment = compute_cache_key(None)
    headers = json.dumps({"Render-Environment": render_environment})

    return SVG_PLACEHOLDER_MARKUP.format(url=url, headers=headers)


def render_diagram(svg: str, title: str, /) -> markupsafe.Markup:
    svg = markupsafe.Markup(
        "data:image/svg+xml;base64,"
        + base64.standard_b64encode(svg.encode("utf-8")).decode("ascii")
    )
    return SVG_WRAP_MARKUP.format(svg_data=svg, title=title)


def load_templates() -> None:
    for idx_path in sorted(c.TEMPLATES_DIR.glob("**/*.yaml")):
        category = re.sub(r"^[A-Za-z0-9]{2}-", "", idx_path.parent.name)
        template_defs = yaml.safe_load(idx_path.read_text(encoding="utf8"))
        for template_def in template_defs:
            template_def["category"] = category
            template_def["path"] = str(
                idx_path.parent / template_def["template"]
            )
            template = Template(**template_def)
            state.templates.append(template)
        _register_template_category(category)


def compute_cache_key(template: Template | None, /) -> str:
    data = {
        "model-explorer-version": capella_model_explorer.__version__,
        "capellambse-version": capellambse.__version__,
        "ctx-diags-version": capellambse_context_diagrams.__version__,
    }
    if template is not None:
        data["template-hash"] = core.compute_file_hash(str(template.path))
    if model_revision := state.model.info.resources["\x00"].rev_hash:
        data["model-revision"] = model_revision
    else:
        data["launch-id"] = c.LAUNCH_ID
    return json.dumps(data)


def process_html_with_toc(html_content: str) -> tuple[str, list[dict]]:
    """Process HTML to extract TOC and inject IDs in a single pass."""
    if not html_content.strip():
        return html_content, []

    try:
        doc = lxml_html.fromstring(html_content)
    except etree.XMLSyntaxError:
        try:
            doc = lxml_html.fragment_fromstring(
                html_content, create_parent=True
            )
        except Exception:
            logger.warning("Cannot parse template result for TOC processing")
            return html_content, []

    toc_items: list[dict[str, t.Any]] = []
    heading_counter: dict[str, int] = {}

    headings = doc.xpath(".//h1 | .//h2 | .//h3 | .//h4 | .//h5 | .//h6")
    assert isinstance(headings, list)

    for heading in headings:
        if not isinstance(heading, lxml_html.HtmlElement):
            continue
        text = heading.text_content().strip()
        if not text:
            continue

        existing_id = heading.get("id", "")

        if not existing_id:
            slug = re.sub(r"[^\w\s-]", "", text.lower())
            slug = re.sub(r"[-\s]+", "-", slug).strip("-")

            if not slug:
                slug = f"heading-{len(toc_items)}"

            count = heading_counter.get(slug, 0)
            heading_counter[slug] = count + 1

            if count > 0:
                slug = f"{slug}-{count}"

            existing_id = slug
            heading.set("id", existing_id)

        assert isinstance(heading.tag, str)
        toc_items.append(
            {
                "level": int(heading.tag[1:]),
                "text": text,
                "id": existing_id,
            }
        )

    processed_html = lxml_html.tostring(doc, encoding="unicode", method="html")
    return processed_html, toc_items



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import time
import typing as t

import prometheus_client

if t.TYPE_CHECKING:
    import capellambse
    import jinja2

    from capella_model_explorer import reports

jinja_env: jinja2.Environment


idle_time_gauge = prometheus_client.Gauge(
    "idletime_minutes",
    "Time in minutes since the last user interaction",
)
last_interaction = time.time()
model: capellambse.MelodyModel

show_uuids: bool = False

templates: list[reports.Template] = []
template_categories: list[reports.TemplateCategory] = []



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import base64
import contextlib
import copy
import hashlib
import logging
import pathlib

import logfmter

ACCESS_LOGGER = "uvicorn.access"


def compute_file_hash(file_path: str):
    """Compute a hash for the given file."""
    if not pathlib.Path(file_path).exists():
        return ""
    hasher = hashlib.blake2b(digest_size=9, usedforsecurity=False)
    with open(file_path, "rb") as f:
        buf = f.read()
        hasher.update(buf)
    return base64.urlsafe_b64encode(hasher.digest()).decode("utf-8")


class Logfmter(logfmter.Logfmter):
    def format(self, record: logging.LogRecord) -> str:
        record = copy.copy(record)

        if record.name == ACCESS_LOGGER:
            assert record.args is not None
            (
                record.client_addr,
                record.method,
                record.path,
                record.http_version,
                record.status_code,
            ) = record.args
            record.msg = "Request completed"
            record.args = ()

        with contextlib.suppress(AttributeError):
            del record.color_message  # type: ignore[attr-defined]

        return super().format(record)


class SuppressWebsocketNoise(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        return not (
            record.name == "uvicorn.error" and hasattr(record, "websocket")
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""The capella_model_explorer package."""

from __future__ import annotations

from importlib import metadata

try:
    __version__ = metadata.version("capella_model_explorer")
except metadata.PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0+unknown"
del metadata



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import contextlib
import json
import logging
import pathlib
import tempfile
import time
import traceback
import typing as t

import capellambse
import jinja2
import prometheus_client
import starlette
import starlette.middleware
from fasthtml import common as fh
from fasthtml import ft

import capella_model_explorer.constants as c
from capella_model_explorer import components, reports, state

logger = logging.getLogger(__name__)


@contextlib.asynccontextmanager
async def lifespan(_):
    logger.info("Configuration:")
    logger.info("\tRoute prefix: '%s'", c.ROUTE_PREFIX)
    logger.info("\tLive mode: %s", c.LIVE_MODE)
    logger.info("\tHost: '%s'", c.HOST)
    logger.info("\tCapella model: '%s'", c.MODEL)
    logger.info("\tTemplates directory: '%s'", c.TEMPLATES_DIR)

    model_spec = capellambse.loadinfo(c.MODEL)
    logger.info("Loading model from: %s", model_spec["path"])
    state.model = capellambse.MelodyModel(**model_spec)
    logger.info("Loading templates from: %s", c.TEMPLATES_DIR)
    reports.load_templates()
    state.jinja_env = jinja2.Environment(
        autoescape=True,
        loader=jinja2.FileSystemLoader(c.TEMPLATES_DIR),
        lstrip_blocks=True,
        trim_blocks=True,
    )
    state.jinja_env.finalize = reports.finalize
    state.jinja_env.filters["make_href"] = reports.make_href_filter
    state.jinja_env.globals["render_diagram"] = reports.diagram_placeholder
    state.jinja_env.tests["diagram"] = lambda obj: isinstance(
        obj, capellambse.model.AbstractDiagram | capellambse.diagram.Diagram
    )
    state.jinja_env.tests["modelelement"] = lambda obj: isinstance(
        obj, capellambse.model.ModelElement
    )
    yield


class UpdateLastInteractionTimeMiddleware(
    starlette.middleware.base.BaseHTTPMiddleware
):
    async def dispatch(self, request, call_next):
        if request.url.path not in ("/metrics", "/favicon.ico"):
            state.last_interaction = time.time()
        return await call_next(request)


if c.LIVE_MODE:
    _app_cls = fh.FastHTMLWithLiveReload
else:
    _app_cls = fh.FastHTML


app = _app_cls(
    htmx=False,
    hdrs=c.HEADERS,
    key_fname=pathlib.Path(tempfile.gettempdir()) / ".sesskey",
    live=c.LIVE_MODE,
    lifespan=lifespan,
    middleware=[
        starlette.middleware.Middleware(UpdateLastInteractionTimeMiddleware),
    ],
    pico=False,
)
ar = fh.APIRouter(prefix=c.ROUTE_PREFIX)
app.static_route_exts(f"{c.ROUTE_PREFIX}/static", "./static")


@app.get("/metrics")
def metrics() -> t.Any:
    idle_time_minutes = (time.time() - state.last_interaction) / 60
    state.idle_time_gauge.set(idle_time_minutes)
    return fh.Response(
        content=prometheus_client.generate_latest(),
        media_type="text/plain",
    )


@app.get("/")
def prefix_redirect(request) -> t.Any:
    if c.ROUTE_PREFIX:
        return fh.RedirectResponse(url=app.url_path_for("main_home"))
    return home(request)


@ar.get("/", name="main_home")
def home(request) -> t.Any:
    """Show home/ landing page with all reports in categories."""
    page_content = (
        ft.Div(
            components.model_information(),
            components.reports_page(),
            cls="flex flex-col space-y-4 place-items-center mx-auto mb-4",
        ),
        components.bottom_bar(),
    )
    return _maybe_wrap_content(request, None, None, page_content)


@ar.get("/model-object-list")
def model_object_list(
    template_id: str, selected_model_element_uuid: str = "", search: str = ""
) -> t.Any:
    """Return (filtered) list of model elements for active template."""
    template = reports.template_by_id(template_id)
    assert template is not None
    return components.model_elements_list(
        template=template,
        selected_id=selected_model_element_uuid,
        search=search,
    )


@ar.get("/rendered-report")
def rendered_report(template_id: str, model_element_uuid: str = "") -> t.Any:
    """Render and return report.

    Takes the template and model element from the application state and
    returns the rendered template.
    """
    template = reports.template_by_id(template_id)
    assert template is not None
    model_element = (
        state.model.by_uuid(model_element_uuid) if model_element_uuid else None
    )
    template_content = template.path.read_text(encoding="utf8")
    jinja_template = state.jinja_env.from_string(template_content)
    try:
        rendered_template = jinja_template.render(
            object=model_element,
            model=state.model,
            diff_data={},
            object_diff={},
        )
        processed_html, toc_items = reports.process_html_with_toc(
            rendered_template
        )
    except Exception:
        elem_repr = ""
        try:
            if model_element is not None:
                elem_repr = model_element._short_repr_()
        except Exception:
            pass
        logger.exception(
            "Error rendering template %r with object %s",
            template_id,
            elem_repr or repr(model_element_uuid),
        )

        full_traceback = traceback.format_exc()
        return ft.Div(
            ft.Div("Error rendering template:", cls="text-xl"),
            fh.Pre(
                full_traceback,
                cls="text-xs prose dark:prose-invert",
            ),
            cls="dark:text-neutral-100 grow content-center",
        )

    toc_button_script = "document.getElementById('toc-toggle-button').classList.remove('hidden');"
    content = ft.Div(
        ft.Div(
            fh.NotStr(processed_html),
            cls="prose svg-display dark:prose-invert max-w-4xl w-full",
        ),
        components.table_of_contents(toc_items) if toc_items else None,
        ft.Script(
            "document.getElementById('root').classList.remove('h-screen');"
            "document.getElementById('print-button').classList.remove('hidden');"
            + (toc_button_script * bool(toc_items))
        ),
        cls="flex flex-row w-full",
    )
    return (
        content,
        fh.HttpHeader("Cache-Control", f"max-age={c.CACHE_MAX_AGE}"),
        fh.HttpHeader("Vary", "Render-Environment"),
    )


@ar.get("/report/{template_id}")
@ar.get("/report/{template_id}/{model_element_uuid}")
def template_page(
    request: starlette.requests.Request,
    template_id: str,
    model_element_uuid: str | None = None,
    search: str = "",
) -> t.Any:
    template = reports.template_by_id(template_id)
    if template is None:
        template_path = pathlib.Path(template_id.replace("|", "/"))
        content = ft.Div(
            ft.Div("Template not found:", cls="text-xl"),
            fh.Code(template_path),
            cls="dark:text-neutral-100 grow content-center",
        )
        return _maybe_wrap_content(request, None, None, content)

    if template.single:
        placeholder = components.report_placeholder(template, None)
    elif model_element_uuid:
        placeholder = components.report_placeholder(
            template, model_element_uuid
        )
    elif len(template.instances) == 1:
        return fh.RedirectResponse(
            app.url_path_for(
                "template_page",
                template_id=template.id,
                model_element_uuid=template.instances[0]["uuid"],
            )
        )
    else:
        placeholder = components.report_placeholder(None, None)

    if (
        request.headers.get("HX-Request") == "true"
        and request.headers.get("HX-Target") != "root"
    ):
        return (
            components.template_container(placeholder),
            components.template_sidebar(
                template=template,
                selected_model_element_uuid=model_element_uuid,
                search=search,
                oob=True,
            ),
            components.breadcrumbs(template, model_element_uuid, oob=True),
        )

    page_content = ft.Div(
        components.template_sidebar(
            template=template,
            selected_model_element_uuid=model_element_uuid,
            search=search,
        ),
        components.template_container(placeholder),
        id="template-page-content",
        cls=(
            "flex",
            "flex-col-reverse",
            "justify-between",
            "lg:flex-row",
            "min-h-full",
            "print:bg-white",
            "print:min-h-auto",
            "w-full",
        ),
    )

    if request.headers.get("HX-Request") == "true":
        return (
            page_content,
            components.breadcrumbs(template, model_element_uuid, oob=True),
        )

    return components.application_shell(
        page_content,
        template=template,
        element=model_element_uuid,
    )


def _maybe_wrap_content(
    request: starlette.requests.Request,
    template: reports.Template | None,
    element: str | None,
    content: t.Any,
) -> t.Any:
    if request.headers.get("HX-Request") == "true":
        return content, components.breadcrumbs(template, element, oob=True)
    return components.application_shell(
        content, template=template, element=element
    )


@ar.get("/render")
def render_template(
    template_id: str,
    model_element_uuid: str | None = None,
) -> t.Any:
    """Request the rendering of a template.

    Returns a template container with loading spinner that itself (htmx)
    requests the rendered template and replaces itself with the rendered
    template.
    """
    template = reports.template_by_id(template_id)
    return (
        components.report_placeholder(template, model_element_uuid),
        components.breadcrumbs(template, model_element_uuid, oob=True),
    )


@ar.get("/diagram/{parent}/{attr}")
def render_diagram(
    parent: str,
    attr: str,
    params: str = "",
) -> t.Any:
    """Request the rendering of a diagram."""
    if params:
        dec_params = json.loads(params)
    else:
        dec_params = {}
    del params

    try:
        parent_obj = state.model.by_uuid(parent)
    except KeyError:
        return ft.Div(f"Model element not found: {parent}")
    try:
        diag = getattr(parent_obj, attr)
    except AttributeError:
        return ft.Div(f"Model element does not have attribute {attr!r}")
    if not isinstance(diag, capellambse.model.AbstractDiagram):
        return ft.Div(f"Attribute {attr!r} is not a diagram")

    try:
        rendered = reports.render_diagram(
            diag.render("svg", **dec_params), diag.name
        )
    except Exception:
        logger.exception("Error rendering diagram %r on %r", attr, parent)
        full_traceback = traceback.format_exc()
        rendered = reports.SVG_ERROR_MARKUP.format(traceback=full_traceback)

    return (
        fh.NotStr(rendered),
        fh.HttpHeader("Cache-Control", f"max-age={c.CACHE_MAX_AGE}"),
        fh.HttpHeader("Vary", "Render-Environment"),
    )


ar.to_app(app)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import time

from starlette import testclient

from capella_model_explorer import app


def _idle_time_minutes(text: str) -> float:
    idle_time_minutes_line = next(
        line
        for line in text.split("\n")
        if line.startswith("idletime_minutes")
    )
    return float(idle_time_minutes_line.split(" ")[1])


def test_idle_time_metric():
    client = testclient.TestClient(app.app)
    response = client.get("/metrics")
    assert response.status_code == 200
    assert "idletime_minutes" in response.text
    idle_time_minutes_1 = _idle_time_minutes(response.text)
    time.sleep(1)
    response = client.get("/metrics")
    idle_time_minutes_2 = _idle_time_minutes(response.text)
    assert idle_time_minutes_2 > idle_time_minutes_1
    client.close()
